# (C) Copyright 2008 Brian Fisher, Zach Young, James Paige
# This is Free Software released under the terms of the GPL
# Created for the Pyweek May 2008 competition: "ROBOTS!"

import os
from random import randint, choice
import math
import copy

from locals import *
from screenthing import ScreenThing
from slantedrect import SlantedRect

#------------------------------------------------------------------------------

class CityStyle(AssetBase):
  pavement_color = AssetColor([0.2, 0.2, 0.2, 1.0], "color of the pavement on the streets")
  road_dash = AssetRef("road_dash")
  max_building_height_hack = AssetInteger(600, "pixels", 0, 10000, "for screen visibility")

#------------------------------------------------------------------------------

class CityMap(object):

  style = AssetRef(CityStyle, "!")

  def __init__(self):
    self.style = "city"
    self.blocks = BlockDict()
    self.blocks.set([0,0], "central_park")

  def draw_ground(self, display, bottom_left, top_right):
    display.erase(color=self.style.pavement_color)
    for pos in get_block_pos_between(bottom_left, top_right):
      block = self.blocks[pos]
      block.draw(display, pos)

  def get_buildings_intersecting(self, bottom_left, top_right):
    list = []
    for pos in get_block_pos_between(bottom_left, top_right):
      block = self.blocks[pos]
      block.get_buildings(list, pos)
    return list
        
#------------------------------------------------------------------------------

class BlockDict(object):

  def __init__(self):
    self._blocks = {}

  def __getitem__(self, index):
    key = (index[0], index[1])
    if not self._blocks.has_key(key):
      # choose a random city block to fill any empty block
      block_asset_id = choice(AvailableAssetIDs(CityBlock))
      self.set(key, block_asset_id)  
    return self._blocks[key]

  def set(self, index, block_asset_id):
    key = (index[0], index[1])
    block_copy = copy.deepcopy(Disgorge(CityBlock, block_asset_id))
    block_copy.original_asset_id = block_asset_id
    self._blocks[key] = block_copy

  def __setitem__(self, index, block):
    self._blocks[(index[0], index[1])] = block
    
  def __iter__(self):
      return self._blocks.__iter__()

#------------------------------------------------------------------------------
def get_block_pos_between(bottom_left, top_right):
  block_end_origin_y = -CityBlock.size[1]/2.0 - CityBlock.street_size[1]
  block_spacing_y = CityBlock.size[1] + CityBlock.street_size[1]
  starting_row = int(math.floor((bottom_left[1] - block_end_origin_y)/block_spacing_y))
  block_start_origin_y = -CityBlock.size[1]/2.0
  ending_row = int(math.floor((top_right[1] - block_start_origin_y)/block_spacing_y))
  
  slanty_width = CityBlock.size[0] + CityBlock.size[1]
  block_spacing_x = CityBlock.size[0] + CityBlock.street_size[0]
  block_end_origin_x = -block_spacing_x + slanty_width/2.0
  block_start_origin_x = -slanty_width/2.0

  positions = []
  if ending_row >= starting_row:
      for row in xrange(starting_row, ending_row + 1):
        this_row_end_origin_x = -block_spacing_x + slanty_width/2.0
        this_row_start_origin_x = -slanty_width/2.0
        if row % 2 == 1:
          this_row_end_origin_x -= block_spacing_x/2.0
          this_row_start_origin_x -= block_spacing_x/2.0
        starting_col = int(math.floor((bottom_left[0] - this_row_end_origin_x)/block_spacing_x))
        ending_col = int(math.floor((top_right[0] - this_row_start_origin_x)/block_spacing_x))
        for col in xrange(starting_col, ending_col + 1):
            positions.append(vec2d(col, row))
  return positions
    
#------------------------------------------------------------------------------

class BuildingStyle(AssetBase):
  image = AssetRef(ImageAsset, "!")
  obstruct_pos = AssetVec2d((0,0), "Position of the building's obstruction box")
  obstruct_size = AssetVec2d((60,30), "Size of the buildings obstruction box")
  can_collide = AssetFlag(True, "Whether or not you can smash the building")
  health = AssetInteger(100, "hit points")
  value = AssetFloat(500000.00, "Dollar value in 1982 US west coast real-estate market")

#------------------------------------------------------------------------------

class Building(AssetBase):

  style = AssetRef(BuildingStyle, "!")
  pos = AssetVec2d()

  def __init__(self):
    self.asset_id = None # deferring the asset id loading is a necesary hack. I don't know why.
    self.smash_sounds = []
    for i in xrange(1,4):
      self.smash_sounds.append(Disgorge(SoundAsset, 'sounds/building_destruct%d' % (i)))

  def still_exists(self):
    if self.asset_id == None:
      # uninitualized damage, therefore not destroyed
      return True
    return self.health > 0

  def draw(self, display, block_pixel_pos):
    display.render(self.style.image, block_pixel_pos + self.pos)

  def damage(self):
    game = HamsterWheel.the_game
    if self.asset_id == None:
      self.asset_id = self.style.asset_metadata["path"]
      self.max_health = self.style.health
      self.health = self.style.health
      self.value = self.style.value
      self.frame_number = 0
    self.health = max(0, self.health - 1)
    if self.health < int(self.max_health * .75):
      damage_frame_number = 3 - int(math.ceil(self.health / (self.max_health * .25)))
      damage_frame = self.asset_id + "_hit" + str(damage_frame_number)
      if damage_frame_number != self.frame_number:
        self.frame_number = damage_frame_number
        if(os.path.exists('data/' + damage_frame + '.png')):
          keep_obstruct_pos = self.style.obstruct_pos
          keep_obstruct_size = self.style.obstruct_size
          self.style = damage_frame
          self.style.obstruct_pos = keep_obstruct_pos
          self.style.obstruct_size = keep_obstruct_size
        game.score_damage += self.value * .25
        if len(self.smash_sounds):
          smash_sound = self.smash_sounds[randint(0, len(self.smash_sounds) - 1)]
          smash_sound.play()

#------------------------------------------------------------------------------

class CityBlock(AssetBase):

  size = vec2d(400, 200)
  street_size = vec2d(100, 50)
  buildings = AssetList(AssetObject(Building))
  color = AssetColor([0.8, 0.8, 0.8, 1.0], "color of the block")
  
  #: this is a lazy way of not having to calculate the height of the tallest
  #: building in the block (which we may want to actually do later on)
  height = 600 #...except that it doesn't work at all :(

  def __init__(self):
    self.rect = SlantedRect(self.size)

  def pixel_pos(self, block_pos):
    block_offset = self.size + self.street_size
    result = block_pos * block_offset
    if block_pos.y % 2 == 1:
      result.x -= block_offset.x / 2
    return result

  def is_onscreen(self, block_pos):
    game = HamsterWheel.the_game
    for corner in self.outer_corners(block_pos):
      if corner.x < game.screen_width and corner.x >= 0 and corner.y < game.screen_height and corner.y >= 0:
        return True
    return False

  def corners(self, block_pos):
    """Returns the corners of a block, relative to the screen (not the map)"""
    pixel_pos = self.pixel_pos(block_pos)
    return self.corners_at(pixel_pos)

  def corners_at(self, pixel_pos):
    game = HamsterWheel.the_game
    return self.rect.corners(pixel_pos - game.camera)

  def outer_corners(self, block_pos):
    pixel_pos = self.pixel_pos(block_pos)
    return self.outer_corners_at(pixel_pos)

  def outer_corners_at(self, pixel_pos):
    corners = self.corners_at(pixel_pos)
    corners[3].x = corners[0].x
    corners[1].x = corners[2].x
    #corners[2].y += self.height
    #corners[3].y += self.height
    return corners

  def get_buildings(self, list, block_pos):
    """append an unsorted list of container objects for each building
    on the block with position relative to the screen based on the
    current camera scroll position relative to the map"""
    pixel_pos = self.pixel_pos(block_pos)
    self.get_buildings_at(list, pixel_pos)

  def get_buildings_at(self, list, pixel_pos):
    """append an unsorted list of container objects for each building
    on the block with position relative to the screen based on an arbitrary
    center point"""
    game = HamsterWheel.the_game
    for building in self.buildings:
      list.append(ScreenThing(building, building.style.image, pixel_pos - game.camera + building.pos))

  def draw(self, display, block_pos):
    pixel_pos = self.pixel_pos(block_pos)
    self.draw_at(display, pixel_pos)

  def draw_at(self, display, pixel_pos):
    game = HamsterWheel.the_game
    # Display the sidewalk defining the block area
    (bl, br, tr, tl) = self.corners_at(pixel_pos)
    display.quad(bl, br, tr, tl, self.color)

  def asset_report_property_change(self, key, value):
      """this is a hacky hook used to make it so if we do live edit, blocks get updated..."""
      asset_id = self.asset_metadata['path']
      game = HamsterWheel.the_game
      for block_pos in game.map.blocks:
          block = game.map.blocks[block_pos]
          if hasattr(block, "original_asset_id"):
              if block.original_asset_id == asset_id:
                  setattr(block, key, copy.deepcopy(value))
      
#------------------------------------------------------------------------------

__BrokenCityStyle = CityStyle()
RegisterAssetExtension(CityStyle, "citystyle", __BrokenCityStyle)

__BrokenCityBlock = CityBlock()
RegisterAssetExtension(CityBlock, "cityblock", __BrokenCityBlock)

__BrokenBuildingStyle = BuildingStyle()
RegisterAssetExtension(BuildingStyle, "buildingstyle", __BrokenBuildingStyle)
