# (C) Copyright 2008 Brian Fisher, Zach Young, James Paige
# This is Free Software released under the terms of the GPL
# Created for the Pyweek May 2008 competition: "ROBOTS!"

from locals import *
from slantedrect import SlantedRect
import math

#------------------------------------------------------------------------------
class RobotStyle(AssetBase):
    aiming_cursor = AssetRef(ImageAsset, "laser_sight")
    aiming_charged_laser_glow = AssetRef(ImageAsset, "laser_sight_glow")
    foot = AssetRef(ImageAsset)
    shin = AssetRef(ImageAsset)
    femur = AssetRef(ImageAsset)
    torso = AssetRef(ImageAsset)
    shoulder = AssetRef(ImageAsset)
    forearm = AssetRef(ImageAsset)
    head = AssetRef(ImageAsset)
    claw = AssetRef(ImageAsset)
    drill = AssetRef(ImageAsset)
    collision_radius = AssetVec2d((50, 20), "defines a collision ellipse")
    walk_speed = AssetFloat(120.0, "pixels/s", .0, 1000.0)
    camera_follow_rate = AssetFloat(.15, "%/frame", .0, 1.0)
    camera_center = AssetVec2d((400,260))
    step_width = AssetFloat(45.0, "pixels")
    hud_font = AssetRef(GlyphSet, "hud")
    smaller_hud_font = AssetRef(GlyphSet, "hud_small")
    satisfaction_happy = AssetRef(ImageAsset)
    satisfaction_unhappy = AssetRef(ImageAsset)
    arrow = AssetRef(ImageAsset)
    jerky_destroy_smooth_steps = AssetInteger(3, "frames", 0, 15, "a counter that makes sure robot arms won't jerk about for intermittent collision")
    
    footstep_sound = AssetRef(SoundAsset)
    hit_sound = AssetRef(SoundAsset)
    
#------------------------------------------------------------------------------
class Robot(object):
    style = AssetRef(RobotStyle)
    
    def __init__(self):
        game = HamsterWheel.the_game
        self.pos = vec2d(0,0)
        self.facing_dir = 1
        self.roto_shoulders = RoboRotate(speed=5)
        self.roto_elbows = RoboRotate(speed=2)
        self.waggle_arms = RoboRotate(speed=3, waggle=True)
        self.rotojoints = [self.roto_shoulders, self.roto_elbows, self.waggle_arms]
        self.debug_collisions = []
        self.typer = GlyphTyper(game.display, self.style.hud_font)
        self.typer2 = GlyphTyper(game.display, self.style.smaller_hud_font)
        self.jerky_arm_countdown = 0
        self.hit_sound_countdown = SoundCountdown(20)
        
        self.foot_cycle_pos = .0
        self.foot_cycle_low_hit = self.foot_cycle_pos - math.pi
        self.foot_cycle_high_hit = self.foot_cycle_pos + math.pi

    def update(self, seconds):
        game = HamsterWheel.the_game
        move_dir = vec2d(0,0)
        if game.keys.holding(K_w) or game.keys.holding(K_UP):
            move_dir.y += 1
        if game.keys.holding(K_a) or game.keys.holding(K_LEFT):
            move_dir.x -= 1
        if game.keys.holding(K_s) or game.keys.holding(K_DOWN):
            move_dir.y -= 1
        if game.keys.holding(K_d) or game.keys.holding(K_RIGHT):
            move_dir.x += 1
        move_dir.normalize()

        right_dir = vec2d(1.0, -.5)
        if right_dir.dot(move_dir) > 0:
            self.facing_dir = 1
        elif right_dir.dot(move_dir) < 0:
            self.facing_dir = -1

        proposed_move = move_dir*self.style.walk_speed*seconds
        collisions = self.check_collision_and_correct_movement(proposed_move)
        self.pos += proposed_move
        
        move_walk_distance = proposed_move.length/self.style.step_width
        self.foot_cycle_pos = self.foot_cycle_pos - self.facing_dir*math.pi*move_walk_distance
        if self.foot_cycle_pos < self.foot_cycle_low_hit:
          self.style.footstep_sound.play()
          self.foot_cycle_pos %= 2*math.pi
          self.foot_cycle_high_hit = self.foot_cycle_pos + math.pi*5/180
          self.foot_cycle_low_hit = self.foot_cycle_pos - math.pi
        if self.foot_cycle_pos > self.foot_cycle_high_hit:
          self.style.footstep_sound.play()
          self.foot_cycle_pos %= 2*math.pi
          self.foot_cycle_low_hit = self.foot_cycle_pos - math.pi*5/180
          self.foot_cycle_high_hit = self.foot_cycle_pos + math.pi
        
        should_smash = False
        if len(collisions) > 0:
            should_smash = True
            self.jerky_arm_countdown = self.style.jerky_destroy_smooth_steps
        elif self.jerky_arm_countdown > 0:
            self.jerky_arm_countdown -= 1
            should_smash = True
            
        if not should_smash:
          # not smashing
          game.debug_text = ""
          self.roto_shoulders.to = -10
          self.roto_elbows.to = 45
          self.waggle_arms.stop()
        else:
          # smash!
          self.roto_shoulders.to = 55
          self.roto_elbows.to = 0
          self.waggle_arms.start(30)
          for building in collisions:
            building.damage()
            if self.hit_sound_countdown():
              self.style.hit_sound.play()
            
        off_camera = self.pos - game.camera - self.style.camera_center
        game.camera += off_camera*self.style.camera_follow_rate
        
        # update body part rotation
        for roto in self.rotojoints:
          roto.update()
        
        # HACK!
        for pos in game.mouse.left_clicks:
          effect = Effect(pos + game.camera, 0, "effects/fire")
          game.effects.append(effect)
    
    def z_sort_order(self):
        game = HamsterWheel.the_game
        return self.pos.y - game.camera.y
    
    def get_joint(self, image, joint_name, pos, angle=None, scale=None):
        try:
            joint_pos = image.joints[joint_name]
        except:
            return pos
        if scale != None:
            joint_pos = joint_pos*scale
        if angle != None:
            joint_pos = joint_pos.rotated(angle)
        joint_pos += pos
        return joint_pos
    
    def draw(self, display):
        game = HamsterWheel.the_game
        draw_pos = self.pos - game.camera 
        
        leg_distance = self.style.shin.joints[0] + self.style.femur.joints[0]
        crotch_pos = (self.style.torso.joints[0] + self.style.torso.joints[1])/2.0
        torso_pos = draw_pos - leg_distance - crotch_pos
        
        direction = vec2d(self.facing_dir, 1)
        
        front_shoulder_angle = (self.roto_shoulders.angle - self.waggle_arms.angle)*self.facing_dir
        back_shoulder_angle = (self.roto_shoulders.angle + self.waggle_arms.angle)*self.facing_dir
        front_arm_angle = (self.roto_shoulders.angle + self.roto_elbows.angle - self.waggle_arms.angle)*self.facing_dir
        back_arm_angle = (self.roto_shoulders.angle + self.roto_elbows.angle + self.waggle_arms.angle)*self.facing_dir
        
        front_foot_offset = vec2d(self.style.step_width/2.0, 0)
        front_foot_offset.rotate(math.degrees(self.foot_cycle_pos))
        back_foot_offset = -front_foot_offset
        front_foot_offset.y = max(front_foot_offset.y, .0)
        back_foot_offset.y = max(back_foot_offset.y, .0)        
        
        front_leg_radians = math.atan2(front_foot_offset.x, leg_distance.length - front_foot_offset.y)
        back_leg_radians = math.atan2(back_foot_offset.x, leg_distance.length - back_foot_offset.y)
        
        front_bent_length = (leg_distance + front_foot_offset).length
        back_bent_length = (leg_distance + back_foot_offset).length
        femur_length = self.style.femur.joints[0].length
        shin_length = self.style.shin.joints[0].length
        
        front_knee_cosine = (femur_length**2 + shin_length**2 - front_bent_length**2)/(2*femur_length*shin_length)
        front_knee_cosine = max(min(1.0, front_knee_cosine), -1.0)
        front_knee_radians = math.acos(front_knee_cosine)

        front_hip_bend = math.asin(shin_length*math.sin(front_knee_radians)/front_bent_length)
        front_hip_radians = front_leg_radians + front_hip_bend*self.facing_dir
        front_knee_radians = front_hip_radians - (math.pi - front_knee_radians)*self.facing_dir
        front_hip_angle = math.degrees(front_hip_radians)
        front_knee_angle = math.degrees(front_knee_radians)
                                    
        back_knee_cosine = (femur_length**2 + shin_length**2 - back_bent_length**2)/(2*femur_length*shin_length)
        back_knee_cosine = max(min(1.0, back_knee_cosine), -1.0)
        back_knee_radians = math.acos(back_knee_cosine)

        back_hip_bend = math.asin(shin_length*math.sin(back_knee_radians)/back_bent_length)
        back_hip_radians = back_leg_radians + back_hip_bend*self.facing_dir
        back_knee_radians = back_hip_radians - (math.pi - back_knee_radians)*self.facing_dir
        back_hip_angle = math.degrees(back_hip_radians)
        back_knee_angle = math.degrees(back_knee_radians)        
        
        front_shoulder_pos = self.get_joint(self.style.torso, 2, torso_pos, 0, direction)
        back_shoulder_pos = self.get_joint(self.style.torso, 3, torso_pos, 0, direction)
        front_forearm_pos = self.get_joint(self.style.shoulder, 0, front_shoulder_pos, front_shoulder_angle, direction)
        back_forearm_pos = self.get_joint(self.style.shoulder, 0, back_shoulder_pos, back_shoulder_angle, direction)
        front_hand_pos = self.get_joint(self.style.forearm, 0, front_forearm_pos, front_arm_angle, direction)
        back_hand_pos = self.get_joint(self.style.forearm, 0, back_forearm_pos, back_arm_angle, direction)

        front_femur_pos = self.get_joint(self.style.torso, 0, torso_pos, 0, direction)
        back_femur_pos = self.get_joint(self.style.torso, 1, torso_pos, 0, direction)
        front_shin_pos = self.get_joint(self.style.femur, 0, front_femur_pos, front_hip_angle, direction)
        back_shin_pos = self.get_joint(self.style.femur, 0, back_femur_pos, back_hip_angle, direction)
        front_foot_pos = self.get_joint(self.style.shin, 0, front_shin_pos, front_knee_angle, direction)
        back_foot_pos = self.get_joint(self.style.shin, 0, back_shin_pos, back_knee_angle, direction)
                
        head_pos = self.get_joint(self.style.torso, 4, torso_pos, 0)

        display.render(self.style.forearm, back_forearm_pos, back_arm_angle, direction)
        display.render(self.style.shoulder, back_shoulder_pos, back_shoulder_angle, direction)
        display.render(self.style.drill, back_hand_pos, back_arm_angle, direction)
        display.render(self.style.foot, back_foot_pos, scale=direction)
        display.render(self.style.shin, back_shin_pos, back_knee_angle, direction)
        display.render(self.style.femur, back_femur_pos, back_hip_angle, direction)
        
        display.render(self.style.torso, torso_pos, scale=direction)
        display.render(self.style.head, head_pos, scale=direction)
        
        display.render(self.style.shin, front_shin_pos, front_knee_angle, direction)
        display.render(self.style.femur, front_femur_pos, front_hip_angle, direction)
        display.render(self.style.foot, front_foot_pos, scale=direction)
        display.render(self.style.forearm, front_forearm_pos, front_arm_angle, direction)
        display.render(self.style.shoulder, front_shoulder_pos, front_shoulder_angle, direction)
        display.render(self.style.claw, front_hand_pos, front_arm_angle, direction)
    
    def draw_ui(self, display):
        game = HamsterWheel.the_game
        mouse = game.mouse
        self.typer.draw("CITY DAMAGE: $%1.2f" % (game.score_damage), [0, game.screen_height - self.style.hud_font.height])
        self.typer2.draw("CURRENT CUSTOMER:", [game.screen_width - 250, game.screen_height - self.style.smaller_hud_font.height])
        self.typer2.draw(game.caller.name, [game.screen_width - 250, game.screen_height - self.style.smaller_hud_font.height * 2])
        self.typer2.draw("HOLD TIME: %s" % (game.caller.time_string()), [game.screen_width - 250, game.screen_height - self.style.smaller_hud_font.height * 3])
        satisfaction_pos = vec2d(game.screen_width - 230, game.screen_height - self.style.smaller_hud_font.height * 4)
        satisfaction = game.caller.satisfaction()
        for i in xrange(0, 5):
          if i < satisfaction:
            display.render(self.style.satisfaction_happy, satisfaction_pos)
          else:
            display.render(self.style.satisfaction_unhappy, satisfaction_pos)
          satisfaction_pos.x += 30
        arrow_pos = [game.screen_width - self.style.arrow.width, 0 + self.style.arrow.height]
        arrow_angle = (game.caller.where_in_pixels() - self.pos).angle
        
        display.render(self.style.arrow, arrow_pos, angle=arrow_angle-90)
        display.render(self.style.aiming_cursor, mouse.pos)
        display.render(self.style.aiming_charged_laser_glow, mouse.pos)
        
    def draw_debug(self, display):
        game = HamsterWheel.the_game
        for collision_dir in self.debug_collisions:
            display.line(((self.pos - game.camera),(self.pos - game.camera - collision_dir*self.style.collision_radius)), (1,1,0,.5))
        self.debug_collisions = []                         
        draw_pos = self.pos - game.camera
        debug_color = (.6, .1, .7, .5)
        display.circle(draw_pos, self.style.collision_radius, debug_color)
        sort_point = vec2d(self.style.camera_center.x, self.z_sort_order())
        display.line((sort_point - (10,0), sort_point + (10, 0)), debug_color) 
        
    def check_collision_and_correct_movement(self, proposed_move):
      game = HamsterWheel.the_game
      new_screen_pos = self.pos + proposed_move - game.camera
      rect = SlantedRect()
      collisions = []
      bottom_left_bounds = self.pos - self.style.collision_radius
      top_right_bounds = self.pos + self.style.collision_radius
      things = game.map.get_buildings_intersecting(bottom_left_bounds, top_right_bounds)
     
      for thing in things:
        if thing.object.still_exists() and thing.object.style.can_collide:
          rect.size = thing.object.style.obstruct_size
          collision_dir = rect.get_ellipse_intersection_dir(new_screen_pos - thing.pos, self.style.collision_radius)
          if collision_dir != None:
            collisions.append(thing.object)
            distance = collision_dir.dot(proposed_move)
            if distance < .0:
              correction = distance * collision_dir
              proposed_move -= correction
              new_screen_pos = self.pos + proposed_move - game.camera
            if game.draw_debug_enabled:
              self.debug_collisions.append(collision_dir)
      return collisions

#------------------------------------------------------------------------------
class RoboRotate(object):
  
  def __init__(self, start_angle=0, speed=1, waggle=False):
    self.angle = start_angle
    self.to = start_angle
    self.speed = speed
    self.waggle = waggle
    self.waggling = False

  def update(self):
    diff = self.to - self.angle
    if abs(diff) <= self.speed:
      self.angle = self.to
      if self.waggle:
        self.to = self.to * -1
    else:
      if diff > 0:
        self.angle += self.speed
      else:
        self.angle -= self.speed

  def start(self, angle):
    if not self.waggling:
      self.to = angle
      self.waggling = True

  def stop(self):
    self.to = 0
    self.waggling = False

#------------------------------------------------------------------------------

class SoundCountdown(object):

  def __init__(self, delay):
    self.delay = delay
    self.counter = delay

  def __call__(self):
    if self.counter == 0:
      self.counter = self.delay
    else:
      self.counter -= 1
      if self.counter == 0:
        return True
    return False

#------------------------------------------------------------------------------
__BrokenRobotStyle = RobotStyle()
RegisterAssetExtension(RobotStyle, "robot", __BrokenRobotStyle)
        